/*
 *  i2c-at24cxx.c
 *
 *  Copyright (C) 2003-2005 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/io.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/ioctl.h>

#include "i2c-at24cxx.h"

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "i2c AT24Cxx driver"
#define LICENSE "GPL v2"

#define CHRDEV  "i2c-at24cxx"
#define VERSION CHRDEV ": i2c at24cxx eeprom driver, (C) 2003-2005 Atmark Techno, Inc."

static struct i2c_driver at24cxx_driver;

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { I2C_EEPROM_ADDRESS, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
  normal_i2c:       normal_addr,
  normal_i2c_range: ignore,
  probe:            ignore,
  probe_range:	    ignore,
  ignore:	    ignore,
  ignore_range:	    ignore,
  force:	    ignore,
};

static int
at24cxx_detect(struct i2c_adapter *adap, int addr, int kind)
{
  struct i2c_client *c;
  
  c = kmalloc(sizeof(*c), GFP_KERNEL);
  if (!c)
    return -ENOMEM;

  memset(c, 0, sizeof(struct i2c_client));
  strcpy(c->name, "AT24Cxx");
  c->flags   = 0;
  c->addr    = addr;
  c->adapter = adap;
  c->driver  = &at24cxx_driver;
  i2c_set_clientdata(c, NULL);
  
  return i2c_attach_client(c);
}

static int
at24cxx_attach(struct i2c_adapter *adap)
{
  return i2c_probe(adap, &addr_data, at24cxx_detect);
}

static int
at24cxx_detach(struct i2c_client *client)
{
  i2c_detach_client(client);
  kfree(client);
  return 0;
}

static int
at24cxx_getbyte(struct i2c_client *client, unsigned char addr)
{
  int ret;
  unsigned char buf[1];
  struct i2c_msg msgs[2] = {
    { client->addr, 0, 1, buf },
    { client->addr, I2C_M_RD, 1, buf },
  };
  
  buf[0] = addr;
  ret = i2c_transfer(client->adapter, msgs, 2);
  if (ret == 2) {
    ret = buf[0];
  }
  else if (ret >= 0) {
    ret = -1;
  }
  
  return ret;
}

static int
at24cxx_setbyte(struct i2c_client *client, unsigned char addr,
		unsigned char data)
{
  int ret;
  unsigned char buf[2];
  struct i2c_msg msgs[1] = {
    { client->addr, 0, 2, buf }
  };
  
  buf[0] = addr;
  buf[1] = data;
  
  ret = i2c_transfer(client->adapter, msgs, 1);
  if (ret == 1) {
    ret = 0;
  }
  else if (ret >= 0) {
    ret = -1;
  }
  
  return ret;
}

static int
at24cxx_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
  int i, val;
  struct eeprom_s *pee = arg;
  int addr = pee->addr;
  
  switch (cmd) {
  case GET_EEPROM_DATA:
    if (addr + pee->count > 256) {
      return -EINVAL;
    }
    for (i = 0; i < pee->count; i++) {
      val = at24cxx_getbyte (client, addr++);
      if (val < 0) {
	return val;
      }
      pee->data[i] = val;
    }
    break;
    
  case SET_EEPROM_DATA:
    if (addr + pee->count > 256) {
      return -EINVAL;
    }
    for (i = 0; i < pee->count; i++) {
      val = at24cxx_setbyte (client, addr++, pee->data[i]);
      if (val < 0) {
	return val;
      }
      mdelay(5);
    }
    break;
    
  default:
    return -ENOIOCTLCMD;
  }
  
  return 0;
}

static struct i2c_driver at24cxx_driver = {
  name:		  "AT24Cxx",
  id:		  I2C_DRIVERID_AT24Cxx,
  flags:	  I2C_DF_NOTIFY,
  attach_adapter: at24cxx_attach,
  detach_client:  at24cxx_detach,
  command:	  at24cxx_command,
};

static int __init at24cxx_init(void)
{
  int ret;
  
  ret = i2c_add_driver(&at24cxx_driver);
  if (ret) {
    printk (KERN_WARNING CHRDEV ": Unable to register %s driver.\n",
	    at24cxx_driver.name);
    return ret;
  }
  
  printk (KERN_INFO VERSION "\n");
  return 0;
}

static void __exit at24cxx_exit(void)
{
  i2c_del_driver(&at24cxx_driver);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(at24cxx_init);
module_exit(at24cxx_exit);
